# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Error reporting.
"""

from __future__ import annotations

import json
import os
from pathlib import Path

from ..data.antsibull_nox_data_util import Level as _DataLevel
from ..data.antsibull_nox_data_util import Location as _DataLocation
from ..data.antsibull_nox_data_util import Message as _DataMessage
from . import Level, Location, Message
from .utils import find_json as _find_json


def parse_pylint_json2_errors(
    *,
    source_path: Path,
    output: str,
) -> list[Message]:
    """
    Parse errors reported by pylint in 'json2' format.
    """
    try:
        data = json.loads(output)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=f"Cannot parse pylint output: {exc}\n{output}",
            )
        ]

    messages = []
    if data["messages"]:
        for message in data["messages"]:
            path = os.path.relpath(message["absolutePath"], source_path)
            messages.append(
                Message(
                    file=path,
                    position=Location(line=message["line"], column=message["column"]),
                    end_position=(
                        Location(
                            line=message["endLine"], column=message.get("endColumn")
                        )
                        if message.get("endLine") is not None
                        else None
                    ),
                    level=Level.ERROR,
                    id=message["messageId"],
                    symbol=message["symbol"],
                    message=message["message"],
                )
            )
    return messages


def parse_ruff_check_errors(
    *,
    source_path: Path,
    output: str,
) -> list[Message]:
    """
    Parse errors reported by ruff check in 'json' format.
    """
    try:
        data = json.loads(output)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=f"Cannot parse ruff check output: {exc}\n{output}",
            )
        ]

    messages = []
    for message in data:
        path = os.path.relpath(message["filename"], source_path)
        hint: str | None = None
        if message.get("fix"):
            fix = message["fix"]
            if "message" in fix:
                hint = fix["message"]
        end_line = message["end_location"]["row"]
        end_col = message["end_location"]["column"] - 1
        if end_col == 0:
            end_line -= 1
            end_col = -1
        messages.append(
            Message(
                file=path,
                position=Location(
                    line=message["location"]["row"],
                    column=message["location"]["column"],
                ),
                end_position=Location(
                    line=end_line,
                    column=end_col,
                ),
                level=Level.ERROR,
                id=message["code"],
                message=message["message"],
                hint=hint,
                url=message["url"],
            )
        )
    return messages


def parse_mypy_errors(
    *,
    root_path: Path,
    source_path: Path,
    output: str,
) -> list[Message]:
    """
    Process errors reported by mypy in 'json' format.
    """
    messages = []
    _mypy_severity = {
        "error": Level.ERROR,
        "note": Level.INFO,
    }

    def plus_one_or_none(value: int | None) -> int | None:
        if value is None:
            return None
        return value + 1

    for line in output.splitlines():
        if not line.strip():
            continue
        try:
            data = json.loads(line)
            path = os.path.relpath(
                root_path / data["file"],
                source_path,
            )
            level = _mypy_severity.get(data["severity"], Level.ERROR)
            messages.append(
                Message(
                    file=path,
                    position=Location(
                        line=data["line"],
                        column=plus_one_or_none(data["column"]),
                    ),
                    end_position=None,
                    level=level,
                    id=data["code"],
                    message=data["message"],
                    hint=data["hint"],
                )
            )
        except Exception:  # pylint: disable=broad-exception-caught
            messages.append(
                Message(
                    file=None,
                    position=None,
                    end_position=None,
                    level=Level.ERROR,
                    id=None,
                    message=f"Cannot parse mypy output: {line}",
                )
            )
    return messages


def parse_bare_framework_errors(
    *,
    output: str,
) -> list[Message]:
    """
    Process errors reported by tools from data with
    antsibull_nox.data.antsibull_nox_data.util.report_result().
    """
    try:
        data = json.loads(output)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=f"Cannot parse output: {exc}\n{output}",
            )
        ]

    def loc(data: _DataLocation | None) -> Location | None:
        if data is None:
            return None
        return Location(line=data.line, column=data.column, exact=data.exact)

    levels: dict[_DataLevel, Level] = {
        "error": Level.ERROR,
        "warning": Level.WARNING,
        "info": Level.INFO,
    }

    messages = []
    for message in data["messages"]:
        msg = _DataMessage.from_json(message)
        messages.append(
            Message(
                file=msg.file,
                position=loc(msg.start),
                end_position=loc(msg.end),
                level=levels.get(msg.level, Level.ERROR),
                id=msg.id,
                message=msg.message,
                symbol=msg.id,
                hint=msg.hint,
                note=msg.note,
                url=msg.url,
            )
        )
    return messages


def parse_antsibull_docs_errors(
    *,
    output: str,
) -> list[Message]:
    """
    Parse errors reported by antsibull-docs lint-collection-docs 'json' format.
    """
    try:
        data = json.loads(_find_json(output))
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=f"Cannot parse antsibull-docs lint-collection-docs output: {exc}\n{output}",
            )
        ]

    messages = []
    for message in data["messages"]:
        row = message.get("row")
        messages.append(
            Message(
                file=message["path"],
                position=(
                    Location(
                        line=row,
                        column=message.get("column"),
                    )
                    if row is not None
                    else None
                ),
                end_position=(
                    Location(
                        line=row,
                        column=message["end_column"],
                    )
                    if row is not None and message.get("end_column") is not None
                    else None
                ),
                level=Level.ERROR,
                id=None,
                message=message["message"],
            )
        )
    return messages


__all__ = (
    "parse_pylint_json2_errors",
    "parse_ruff_check_errors",
    "parse_mypy_errors",
    "parse_bare_framework_errors",
    "parse_antsibull_docs_errors",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2026, Ansible Project

"""
Path matchers.
"""

from __future__ import annotations

import dataclasses
import typing as t
from collections.abc import Iterable
from pathlib import Path

from .utils import path_walk


def _split_path(path: Path) -> tuple[str, ...]:
    """
    Split relative path ``path`` into a list of path segments.
    """
    assert not path.anchor
    parts = []
    while path.name:
        parts.append(path.name)
        path = path.parent
    return tuple(reversed(parts))


@dataclasses.dataclass(frozen=True)
class _FileInfo:
    """
    Contains information for a file.
    """

    path: Path
    is_dir: bool

    @classmethod
    def create(cls, path: Path) -> t.Self:
        """
        Create a file info object for a path.
        """
        return cls(path=path, is_dir=path.is_dir())


@dataclasses.dataclass(frozen=True)
class _FileSet:
    """
    Contains a set of files as a list of string tuples and a dictionary with further information.
    """

    files: list[tuple[str, ...]]
    infos: dict[tuple[str, ...], _FileInfo]

    @classmethod
    def create(cls, paths: Iterable[Path]) -> t.Self:
        """
        Create a file set from a sequence of paths.
        """
        files = []
        infos = {}
        for path in paths:
            file = _split_path(path)
            if file not in infos:
                files.append(file)
                infos[file] = _FileInfo.create(path)
        return cls(
            files=files,
            infos=infos,
        )

    @classmethod
    def create_from_split(cls, *, split_files: Iterable[tuple[str, ...]]) -> t.Self:
        """
        Create a file set from a sequence of split paths.
        """
        files = []
        infos = {}
        root = Path(".")
        for file in split_files:
            if file not in infos:
                files.append(file)
                infos[file] = _FileInfo.create(root.joinpath(*file))
        return cls(
            files=files,
            infos=infos,
        )

    def clone(self) -> _FileSet:
        """
        Create a copy of this file set.
        """
        return _FileSet(
            files=list(self.files),
            infos=dict(self.infos),
        )

    def subset(self, files: set[tuple[str, ...]]) -> _FileSet:
        """
        Restrict a file set to a subset of files.
        """
        return _FileSet(
            files=list(files), infos={file: self.infos[file] for file in files}
        )

    def merge_set(self, other: _FileSet) -> None:
        """
        Merge this file set with another one.
        """
        for file, info in other.infos.items():
            if file not in self.infos:
                self.files.append(file)
                self.infos[file] = info

    def merge_paths(self, *, paths: Iterable[Path]) -> None:
        """
        Merge this file set with a sequence of paths.
        """
        for path in paths:
            file = _split_path(path)
            if file not in self.infos:
                self.files.append(file)
                self.infos[file] = _FileInfo.create(path)

    def get_paths(self) -> list[Path]:
        """
        Return a list of ``Path`` object for all contained files.
        """
        return [info.path for info in self.infos.values()]


class _ExtensionChecker:
    """
    Allows to test filenames for a set of extensions.
    """

    def __init__(self, *, extensions: Iterable[str]) -> None:
        """
        Create an extension checker, given a list of extensions (without leading period).
        """
        self._extensions = list({f".{ext}" for ext in extensions})

    def has(self, filename: str) -> bool:
        """
        Test whether the filename has one of our extensions.
        """
        return any(filename.endswith(ext) for ext in self._extensions)


@dataclasses.dataclass(frozen=False)
class _FileTreeNode:
    """
    Represent a node in a file tree.
    """

    file: tuple[str, ...]
    contained: bool
    children: dict[str, _FileTreeNode]

    def iterate(self) -> t.Generator[tuple[str, ...]]:
        """
        Generate all contained files.
        """
        if self.contained:
            yield self.file
        for child in self.children.values():
            yield from child.iterate()


@dataclasses.dataclass(frozen=False)
class _FileTree:
    """
    Represent a set of files as a hierarchical tree.
    """

    root: _FileTreeNode

    def __init__(self) -> None:
        self.root = _FileTreeNode(file=(), contained=False, children={})

    def add(self, file: tuple[str, ...], *, keep_pruned: bool = False) -> None:
        """
        Add a file to the tree.

        If ``keep_pruned == True``, all children below an entry will be removed.
        """
        node = self.root
        for index, part in enumerate(file):
            if keep_pruned and node.contained:
                return
            next_node = node.children.get(part)
            if next_node is None:
                next_node = _FileTreeNode(
                    file=file[: index + 1], contained=False, children={}
                )
                node.children[part] = next_node
            node = next_node
        node.contained = True
        if keep_pruned:
            node.children.clear()

    @classmethod
    def create_from_files(
        cls, files: Iterable[tuple[str, ...]], *, keep_pruned: bool = False
    ) -> t.Self:
        """
        Create a file tree from a sequence of files.
        """
        result = cls()
        for file in files:
            result.add(file, keep_pruned=keep_pruned)
        return result

    @classmethod
    def create_from_paths(
        cls, paths: Iterable[Path], *, keep_pruned: bool = False
    ) -> t.Self:
        """
        Create a file tree from a sequence of paths.
        """
        result = cls()
        for path in paths:
            file = _split_path(path)
            result.add(file, keep_pruned=keep_pruned)
        return result

    def find_closest(self, file: tuple[str, ...]) -> tuple[tuple[str, ...], bool]:
        """
        Find the closest match to the file in the tree.

        Return a tuple ``(closest_match, closest_match_contained_in_tree)``.
        """
        node = self.root
        for part in file:
            next_node = node.children.get(part)
            if next_node is None:
                break
            node = next_node
        return node.file, node.contained

    def has_or_has_children(self, file: tuple[str, ...]) -> bool:
        """
        Check if a file is part of the tree, or has children inside the tree
        (i.e. files in the tree start with the given file).
        """
        path, _contained = self.find_closest(file)
        return len(path) == len(file)

    def has_or_is_child(self, file: tuple[str, ...]) -> bool:
        """
        Check if a file is part of the tree, or is a child of something in the tree.
        """
        node = self.root
        for part in file:
            if node.contained:
                return True
            next_node = node.children.get(part)
            if next_node is None:
                return False
            node = next_node
        return node.contained

    def iterate(
        self, *, prefix: tuple[str, ...] | None = None
    ) -> t.Generator[tuple[str, ...]]:
        """
        Generate all contained files.
        """
        node = self.root
        if prefix is not None:
            for part in prefix:
                next_node = node.children.get(part)
                if next_node is None:
                    return
                node = next_node
        yield from node.iterate()


class FileCollector:
    """
    Modifies a list of paths by restricting to and/or removing paths.

    The paths can point to directories or files.
    """

    def __init__(self, *, paths: list[Path] | _FileSet) -> None:
        """
        Create a list of paths.
        """
        self._paths = paths if isinstance(paths, _FileSet) else _FileSet.create(paths)

    def clone(self) -> FileCollector:
        """
        Create a copy of the file collector.
        """
        return FileCollector(paths=self._paths.clone())

    @classmethod
    def create(cls, sources: Iterable[str], *, glob: bool = True) -> t.Self:
        """
        Create a list of paths from strings with potential globbing.
        """
        # Collect files in tree
        tree = _FileTree()
        root = Path(".")
        for source in sources:
            if glob and any(ch in "*?[" for ch in source):
                for path in root.glob(source):
                    tree.add(_split_path(path), keep_pruned=True)
            else:
                tree.add(_split_path(Path(source)), keep_pruned=True)
        # Create set
        return cls(paths=_FileSet.create_from_split(split_files=list(tree.iterate())))

    @staticmethod
    def _get_pruned_tree(paths: list[Path] | FileCollector) -> _FileTree:
        if isinstance(paths, FileCollector):
            return _FileTree.create_from_files(
                paths._paths.files, keep_pruned=True  # pylint: disable=protected-access
            )
        return _FileTree.create_from_paths(paths, keep_pruned=True)

    def restrict(self, *, paths: list[Path] | FileCollector) -> None:
        """
        Restrict the list of paths to the given list of paths.
        """
        paths_tree = self._get_pruned_tree(paths)
        files: set[tuple[str, ...]] = set()
        path_files: set[tuple[str, ...]] = set()
        for file in self._paths.files:
            path, contained = paths_tree.find_closest(file)
            if contained:
                files.add(file)
            elif len(path) == len(file):
                path_files.update(paths_tree.iterate(prefix=path))
        self._paths = self._paths.subset(files)
        if path_files:
            self._paths.merge_set(_FileSet.create_from_split(split_files=path_files))

    def _scan_remove_paths(
        self, path: Path, *, remove: _FileTree, extensions: _ExtensionChecker | None
    ) -> list[Path]:
        result = []
        for root, dirs, files in path_walk(path, top_down=True):
            root_file = _split_path(root)
            if remove.has_or_is_child(root_file):
                # This should never happen anyway, since it's already covered by other cases
                dirs[:] = []  # pragma: no cover
                continue  # pragma: no cover
            if not remove.has_or_has_children(root_file):
                dirs[:] = []  # do not iterate deeper
                result.append(root)
                continue
            for file in files:
                if extensions and not extensions.has(file):
                    continue
                file_file = root_file + (file,)
                if not remove.has_or_is_child(file_file):
                    result.append(root / file)
            for directory in list(dirs):
                # We should probably use .gitignore here...
                if directory == "__pycache__":
                    dirs.remove(directory)
                    continue
                directory_file = root_file + (directory,)
                if remove.has_or_is_child(directory_file):
                    dirs.remove(directory)
                    continue
        return result

    def remove(
        self, *, paths: list[Path] | FileCollector, extensions: list[str] | None = None
    ) -> None:
        """
        Restrict/refine the list of paths by removing a given list of paths.

        If ``extensions`` is provided, during refinement only files with extensions
        in the given list are added.
        """
        extensions_checker = (
            _ExtensionChecker(extensions=extensions) if extensions is not None else None
        )
        paths_tree = self._get_pruned_tree(paths)
        files = set()
        other_files = []
        for file, info in self._paths.infos.items():
            path, contained = paths_tree.find_closest(file)
            if contained:
                continue
            if not info.is_dir or len(path) != len(file):
                files.add(file)
                continue
            other_files.extend(
                self._scan_remove_paths(
                    info.path, remove=paths_tree, extensions=extensions_checker
                )
            )
        self._paths = self._paths.subset(files)
        if other_files:
            self._paths.merge_paths(paths=other_files)

    def get_paths(self) -> list[Path]:
        """
        Return the list of paths.
        """
        return self._paths.get_paths()

    def get_existing(self) -> list[Path]:
        """
        Return the list of paths that actually exist.
        """
        return [path for path in self._paths.get_paths() if path.exists()]


__all__ = ("FileCollector",)



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all plugin and module documentation adheres to yamllint."""

from __future__ import annotations

import io
import os
import sys
import traceback

from antsibull_nox_data_util import (  # type: ignore
    Level,
    Location,
    Message,
    report_result,
    setup,
)
from yamllint import linter
from yamllint.cli import find_project_config_filepath
from yamllint.config import YamlLintConfig
from yamllint.linter import PROBLEM_LEVELS

REPORT_LEVELS: dict[PROBLEM_LEVELS, Level] = {
    "warning": "warning",
    "error": "error",
}


def lint(
    *,
    messages: list[Message],
    path: str,
    data: str,
    config: YamlLintConfig,
) -> None:
    try:
        problems = linter.run(
            io.StringIO(data),
            config,
            path,
        )
        for problem in problems:
            level = REPORT_LEVELS.get(problem.level)
            if level is None:
                continue
            msg = f"{problem.level}: {problem.desc}"
            messages.append(
                Message(
                    file=path,
                    start=Location(line=problem.line, column=problem.column),
                    end=None,
                    level=level,
                    id=problem.rule or None,
                    message=msg,
                )
            )
    except Exception as exc:
        error = str(exc).replace("\n", " / ")
        messages.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=f"Internal error while linting YAML: exception {type(exc)}:"
                f" {error}; traceback: {traceback.format_exc()!r}",
            )
        )


def process_yaml_file(
    messages: list[Message],
    path: str,
    config: YamlLintConfig,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as stream:
            data = stream.read()
    except Exception as exc:
        messages.append(
            Message(
                file=path,
                start=None,
                end=None,
                level="error",
                id=None,
                message=f"Error while parsing Python code: exception {type(exc)}:"
                f" {exc}; traceback: {traceback.format_exc()!r}",
            )
        )
        return

    lint(
        messages=messages,
        path=path,
        data=data,
        config=config,
    )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    config: str | None = extra_data.get("config")

    if config is None:
        config = find_project_config_filepath()

    if config:
        yamllint_config = YamlLintConfig(file=config)
    else:
        yamllint_config = YamlLintConfig(content="extends: default")

    messages: list[Message] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        process_yaml_file(messages, path, yamllint_config)

    return report_result(messages)


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

from __future__ import annotations

import pytest

from antsibull_nox.messages.utils import find_json, find_json_line

FIND_JSON_DATA: list[tuple[str, str]] = [
    (
        r"""""",
        r"""""",
    ),
    (
        r"""foo""",
        r"""foo""",
    ),
    (
        r"""
foo
[bar]
baz
""",
        r"""[bar]""",
    ),
    (
        r"""
{foo}
[bar]
{baz}
""",
        r"""{foo}
[bar]
{baz}""",
    ),
    (
        r"""
{foo
[bar]
baz}
""",
        r"""{foo
[bar]
baz}""",
    ),
    (
        r"""
foo {
  {
foo
}
[
]
""",
        r"""  {
foo
}""",
    ),
    (
        r"""{""",
        r"""{""",
    ),
    (
        r"""
foo
{
bar
""",
        r"""
foo
{
bar
""",
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    FIND_JSON_DATA,
)
def test_find_json(output: str, expected_result: str) -> None:
    assert find_json(output) == expected_result


FIND_JSON_LINE_DATA: list[tuple[str, str]] = [
    (
        r"""""",
        r"""""",
    ),
    (
        r"""{foo}""",
        r"""{foo}""",
    ),
    (
        r"""[foo]""",
        r"""[foo]""",
    ),
    (
        r"""
{foo}
[bar]
{baz}
""",
        r"""{foo}""",
    ),
    (
        r"""
{foo
[bar]
baz}
""",
        r"""[bar]""",
    ),
    (
        r"""
bar
[foo
""",
        r"""
bar
[foo
""",
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    FIND_JSON_LINE_DATA,
)
def test_find_json_line(output: str, expected_result: str) -> None:
    assert find_json_line(output) == expected_result



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

"""
Tests for the collection.install module
"""

from __future__ import annotations

from pathlib import Path

import pytest

from antsibull_nox.collection.data import (
    CollectionData,
)
from antsibull_nox.collection.install import (
    _add_all_dependencies,
    _extract_collections_from_extra_deps_file,
    _MissingDependencies,
)
from antsibull_nox.collection.search import (
    CollectionList,
)


def test__add_all_dependencies() -> None:
    path = Path.cwd()
    all_collections = CollectionList.create(
        {
            c.full_name: c
            for c in [
                CollectionData.create(path=path, full_name="foo.bar", current=True),
                CollectionData.create(
                    path=path, full_name="foo.bam", dependencies={"foo.bar": "*"}
                ),
                CollectionData.create(
                    path=path,
                    full_name="foo.baz",
                    dependencies={"foo.bar": "*", "foo.bam": ">= 1.0.0"},
                ),
                CollectionData.create(
                    path=path, full_name="foo.foo", dependencies={"foo.bam": "*"}
                ),
                CollectionData.create(
                    path=path,
                    full_name="foo.error",
                    dependencies={"foo.does_not_exist": "*"},
                ),
            ]
        }
    )

    # No deps
    deps: dict[str, CollectionData] = {}
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps == {}  # pylint: disable=use-implicit-booleaness-not-comparison
    assert missing.is_empty()

    # Collection without deps
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bar",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar"}
    assert missing.is_empty()

    # Collection with single dep
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bam",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam"}
    assert missing.is_empty()

    # Collection with two deps
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.baz",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam", "foo.baz"}
    assert missing.is_empty()

    # Collection with dependency chain where leaf is already there
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bar",
            "foo.foo",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam", "foo.foo"}
    assert missing.is_empty()

    # Missing collection
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.error",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert not missing.is_empty()
    assert missing.get_missing_names() == ["foo.does_not_exist"]


def test__extract_collections_from_extra_deps_file_special(tmp_path: Path) -> None:
    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert _extract_collections_from_extra_deps_file(tmp_path / "does-not-exist") == []

    dir = tmp_path / "test1"
    dir.mkdir()
    with pytest.raises(
        ValueError,
        match="Error while loading collection dependency file.*Is a directory:",
    ):
        _extract_collections_from_extra_deps_file(dir)


EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_DATA: list[tuple[str, list[str]]] = [
    (
        r"""
collections: []
""",
        [],
    ),
    (
        r"""
collections:
    - foo
    - name: bar
""",
        ["foo", "bar"],
    ),
    (
        # Not exactly legal, but works:
        r"""
collections:
    foo: bar
    baz:
""",
        ["foo", "baz"],
    ),
]


@pytest.mark.parametrize(
    "content, expected_result",
    EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_DATA,
)
def test__extract_collections_from_extra_deps_file(
    content: str, expected_result: list[str], tmp_path: Path
) -> None:
    file = tmp_path / "test1"
    file.write_text(content)
    assert _extract_collections_from_extra_deps_file(file) == expected_result


EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_FAIL_DATA: list[tuple[str, str]] = [
    (
        r"""
collections:
    foo: bar
    23: baz
""",
        "Collection entry #2 must be a string or dictionary",
    ),
    (
        r"""
collections:
    - 42
""",
        "Collection entry #1 must be a string or dictionary",
    ),
    (
        r"""
collections:
    - name:
""",
        "Collection entry #1 does not have a 'name' field of type string",
    ),
    (
        r"""
collections:
    - bar: baz
""",
        "Collection entry #1 does not have a 'name' field of type string",
    ),
]


@pytest.mark.parametrize(
    "content, expected_match",
    EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_FAIL_DATA,
)
def test__extract_collections_from_extra_deps_file_fail(
    content: str, expected_match: str, tmp_path: Path
) -> None:
    file = tmp_path / "test1"
    file.write_text(content)
    with pytest.raises(ValueError, match=expected_match):
        _extract_collections_from_extra_deps_file(file)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox sessions.
"""

from __future__ import annotations

from .collections import (  # Re-export for usage in noxfiles
    CollectionSetup,
    prepare_collections,
)

__all__ = [
    "CollectionSetup",
    "prepare_collections",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Path utils for creating nox sessions.
"""

from __future__ import annotations

import typing as t
from collections.abc import Sequence
from pathlib import Path

from ...cd import get_changes
from ...config import CONFIG_FILENAME
from ...paths.match import FileCollector
from ...python.python_dependencies import get_python_dependency_info

PythonDependencies = t.Literal["none", "imported-by-changed", "importing-changed"]


def add_python_deps(files: list[Path], *, forward: bool, cwd: Path) -> None:
    """
    Given a list of files, add all Python files that (transitively) import
    these (``forward == False``) or are imported by them (``forward == True``).
    """
    deps_info = get_python_dependency_info()
    next_modules_dict = (
        deps_info.file_to_imported_modules
        if forward
        else deps_info.file_to_imported_by_modules
    )
    added = {
        (file if file.is_absolute() else cwd / file)
        for file in files
        if file.name.endswith(".py")
    }
    processed = set()
    to_process = set(added)
    while to_process:
        elt = to_process.pop()
        if elt in processed:
            continue
        processed.add(elt)
        if elt not in added:
            if elt.is_relative_to(cwd):
                files.append(elt.relative_to(cwd))
                added.add(elt)
        next_module_paths = next_modules_dict.get(elt)
        if next_module_paths is None:
            continue
        for next_module in next_module_paths[1]:
            if next_module not in added and next_module not in processed:
                to_process.add(next_module)


def _massage_paths_to_trigger_full_build(
    paths_to_trigger_full_build: Sequence[Path] | None,
) -> list[Path]:
    if paths_to_trigger_full_build is None:
        paths_to_trigger_full_build = []
    paths_to_trigger_full_build = list(paths_to_trigger_full_build)
    paths_to_trigger_full_build.append(Path(CONFIG_FILENAME))
    return paths_to_trigger_full_build


def filter_paths(
    paths: list[Path] | FileCollector,
    /,
    remove: list[Path] | FileCollector | None = None,
    restrict: list[Path] | FileCollector | None = None,
    extensions: list[str] | None = None,
    with_cd: bool = False,
    cd_add_python_deps: PythonDependencies = "none",
    paths_to_trigger_full_build: Sequence[Path] | None = None,
) -> list[Path]:
    """
    Modifies a list of paths by restricting to and/or removing paths.
    """
    paths_to_trigger_full_build = _massage_paths_to_trigger_full_build(
        paths_to_trigger_full_build
    )
    collector = (
        paths.clone()
        if isinstance(paths, FileCollector)
        else FileCollector(paths=[Path(path) for path in paths])
    )
    if with_cd:
        cwd = Path.cwd()
        changed_files = get_changes(relative_to=cwd)
        if changed_files is not None and (
            all(file not in paths_to_trigger_full_build for file in changed_files)
        ):
            if cd_add_python_deps != "none":
                add_python_deps(
                    changed_files,
                    forward=cd_add_python_deps == "imported-by-changed",
                    cwd=cwd,
                )
            if extensions:
                changed_files = [
                    file for file in changed_files if file.suffix in extensions
                ]
            collector.restrict(paths=changed_files)
    if restrict:
        collector.restrict(paths=restrict)
    if remove:
        collector.remove(paths=remove, extensions=extensions)
    return collector.get_existing()


def filter_files_cd(
    files: Sequence[Path], *, paths_to_trigger_full_build: Sequence[Path] | None = None
) -> list[Path]:
    """
    Given a sequence of paths, filters out changed files if change detection is enabled.
    If it is disabled, simply return the sequence as a list.
    """
    paths_to_trigger_full_build = _massage_paths_to_trigger_full_build(
        paths_to_trigger_full_build
    )
    changed_files = get_changes(relative_to=Path.cwd())
    if changed_files is None or (
        any(file in paths_to_trigger_full_build for file in changed_files)
    ):
        if isinstance(files, list):
            return files
        return list(files)

    changed_set = set(changed_files)
    return [file for file in files if file in changed_set]


__all__ = [
    "add_python_deps",
    "filter_files_cd",
    "filter_paths",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utils for dealing with packages.
"""

from __future__ import annotations

import dataclasses
import os
import typing as t
from collections.abc import Iterator, Sequence

import nox

from . import IN_CI as _IN_CI

ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not _IN_CI)).lower() in (
    "1",
    "true",
)


@dataclasses.dataclass
class PackageName:
    """
    A PyPI package name.
    """

    name: str

    def get_pip_install_args(self) -> Iterator[str]:
        """
        Yield arguments to 'pip install'.
        """
        yield self.name


@dataclasses.dataclass
class PackageEditable:
    """
    A PyPI package name that should be installed editably (if allowed).
    """

    name: str

    def get_pip_install_args(self) -> Iterator[str]:
        """
        Yield arguments to 'pip install'.
        """
        # Don't install in editable mode in CI or if it's explicitly disabled.
        # This ensures that the wheel contains all of the correct files.
        if ALLOW_EDITABLE:
            yield "-e"
        yield self.name


@dataclasses.dataclass
class PackageRequirements:
    """
    A Python requirements.txt file.
    """

    file: str

    def get_pip_install_args(self) -> Iterator[str]:
        """
        Yield arguments to 'pip install'.
        """
        yield "-r"
        yield self.file


@dataclasses.dataclass
class PackageConstraints:
    """
    A Python constraints.txt file.
    """

    file: str

    def get_pip_install_args(self) -> Iterator[str]:
        """
        Yield arguments to 'pip install'.
        """
        yield "-c"
        yield self.file


PackageType = t.Union[
    str,
    PackageName,
    PackageEditable,
    PackageRequirements,
    PackageConstraints,
]

PackageTypeOrList = t.Union[PackageType, Sequence[PackageType]]


def normalize_package_type(packages: PackageTypeOrList | None) -> list[PackageType]:
    """
    Given a package type or a list of package types or ``None``,
    convert this to a list of package types.
    """
    if packages is None:
        return []
    # A str is a Sequence as well.
    if isinstance(packages, Sequence) and not isinstance(packages, str):
        return list(packages)  # return a copy
    return [packages]


def _get_install_params(packages: Sequence[PackageType]) -> list[str]:
    new_args: list[str] = []
    for arg in packages:
        if isinstance(arg, str):
            new_args.append(arg)
        else:
            new_args.extend(arg.get_pip_install_args())
    return new_args


def install(session: nox.Session, *args: PackageType, **kwargs):
    """
    Install Python packages.
    """
    if not args:
        return

    # nox --no-venv
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn(f"No venv. Skipping installation of {args}")
        return

    new_args = _get_install_params(args)
    session.install(*new_args, "-U", **kwargs)


def check_package_types(
    session: nox.Session, arg_name: str, packages: list[PackageType]
) -> list[PackageType]:
    """
    Given a list of packages, check for invalid package names.
    """
    for package in packages:
        if not isinstance(package, (PackageName, PackageEditable)):
            continue
        if package.name.startswith("-"):
            session.warn(
                f"DEPRECATION WARNING: {arg_name} contains a package name"
                f" {package.name!r} starting with a dash."
                " This behavior is deprecated and will stop working in a future release."
            )
    return packages


__all__ = [
    "PackageName",
    "PackageEditable",
    "PackageRequirements",
    "PackageType",
    "check_package_types",
    "install",
    "normalize_package_type",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

import re
from pathlib import Path

import pytest
from antsibull_fileutils.yaml import store_yaml_file

from antsibull_nox.ansible_test_config import (
    AnsibleTestConfig,
    ModulesConfig,
    get_min_python_version,
    load_ansible_test_config,
)
from antsibull_nox.utils import Version

from .utils import chdir


def test_load_ansible_test_config(tmp_path: Path) -> None:
    tests = tmp_path / "tests"
    tests.mkdir()
    config_yml = tests / "config.yml"
    with chdir(tmp_path):
        with pytest.raises(FileNotFoundError):
            load_ansible_test_config()

        config = load_ansible_test_config(accept_missing=True)
        assert config.modules.python_requires == "default"
        assert config.modules == ModulesConfig()
        assert config == AnsibleTestConfig()

        config = load_ansible_test_config(ignore_errors=True)
        assert config.modules.python_requires == "default"
        assert config.modules == ModulesConfig()
        assert config == AnsibleTestConfig()

        store_yaml_file(config_yml, {"modules": {"python_requires": []}})
        with pytest.raises(ValueError):
            load_ansible_test_config()
        with pytest.raises(ValueError):
            load_ansible_test_config(accept_missing=True)
        config = load_ansible_test_config(ignore_errors=True)
        assert config.modules.python_requires == "default"
        assert config.modules == ModulesConfig()
        assert config == AnsibleTestConfig()

        store_yaml_file(config_yml, {"modules": {"python_requires": "asdf"}})
        config = load_ansible_test_config()
        assert config.modules.python_requires == "asdf"


def test_get_min_python_version() -> None:
    assert (
        get_min_python_version(
            AnsibleTestConfig(modules=ModulesConfig(python_requires="default"))
        )
        == "default"
    )
    assert (
        get_min_python_version(
            AnsibleTestConfig(modules=ModulesConfig(python_requires="controller"))
        )
        == "controller"
    )

    with pytest.raises(
        ValueError,
        match="^%s$"
        % re.escape("Invalid specifier set in modules.python_requires: 'asdf'"),
    ):
        get_min_python_version(
            AnsibleTestConfig(modules=ModulesConfig(python_requires="asdf"))
        )
    assert (
        get_min_python_version(
            AnsibleTestConfig(modules=ModulesConfig(python_requires="asdf")),
            ignore_errors=True,
        )
        == "default"
    )

    matcher = get_min_python_version(
        AnsibleTestConfig(modules=ModulesConfig(python_requires=">=3.8"))
    )
    assert callable(matcher)
    assert not matcher(Version.parse("2.6"))
    assert not matcher(Version.parse("2.7"))
    assert not matcher(Version.parse("3.4"))
    assert not matcher(Version.parse("3.5"))
    assert not matcher(Version.parse("3.6"))
    assert not matcher(Version.parse("3.7"))
    assert matcher(Version.parse("3.8"))
    assert matcher(Version.parse("3.9"))
    assert matcher(Version.parse("3.10"))
    assert matcher(Version.parse("3.11"))
    assert matcher(Version.parse("3.12"))

    matcher = get_min_python_version(
        AnsibleTestConfig(
            modules=ModulesConfig(python_requires=">=3.8,<3.11,!=3.9,!=3.8.1")
        )
    )
    assert callable(matcher)
    assert not matcher(Version.parse("2.6"))
    assert not matcher(Version.parse("2.7"))
    assert not matcher(Version.parse("3.4"))
    assert not matcher(Version.parse("3.5"))
    assert not matcher(Version.parse("3.6"))
    assert not matcher(Version.parse("3.7"))
    assert matcher(Version.parse("3.8"))
    assert not matcher(Version.parse("3.9"))
    assert matcher(Version.parse("3.10"))
    assert not matcher(Version.parse("3.11"))
    assert not matcher(Version.parse("3.12"))



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# PYTHON_ARGCOMPLETE_OK

"""Entrypoint to the antsibull-docs script."""

from __future__ import annotations

import argparse
import os
import os.path
import sys
from collections.abc import Callable
from pathlib import Path

from . import __version__
from .cd import get_base_branch, get_changes, init_cd, supports_cd
from .config import CONFIG_FILENAME, load_config_from_toml
from .init import create_initial_config as _create_initial_config
from .lint_config import lint_config as _lint_config
from .sessions.utils.paths import PythonDependencies, add_python_deps

try:
    import argcomplete

    HAS_ARGCOMPLETE = True
except ImportError:
    HAS_ARGCOMPLETE = False


def lint_config(_: argparse.Namespace) -> int:
    """
    Lint antsibull-nox config file.
    """
    errors = _lint_config()
    for error in errors:
        print(error)
    return 0 if len(errors) == 0 else 3


def create_initial_config(_: argparse.Namespace) -> int:
    """
    Create noxfile.py and antsibull-nox.toml.
    """
    try:
        _create_initial_config()
    except Exception as exc:  # pylint: disable=broad-exception-caught
        print(f"Error: {exc}", file=sys.stderr)
        return 3
    return 0


def show_changes(args: argparse.Namespace) -> int:
    """
    Show changes.
    """
    add_python_deps_opt: PythonDependencies = args.add_python_deps

    config_path = Path(CONFIG_FILENAME)
    cwd = Path.cwd()
    try:
        config = load_config_from_toml(config_path)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        print(f"Error: {exc}", file=sys.stderr)
        return 3

    try:
        init_cd(config=config, config_path=config_path, force=True)
        if not supports_cd():
            print(f"{config_path} does not support change detection.", file=sys.stderr)
            return 3

        base_branch = get_base_branch()
        changes = get_changes(relative_to=cwd) or []
    except ValueError as exc:
        print(f"Error while fetching changes: {exc}", file=sys.stderr)
        return 3

    original = set(changes)
    python_deps: set[Path] = set()
    python_appendum = ""

    changes_ex = list(changes)
    if add_python_deps_opt != "none":
        forward = add_python_deps_opt == "imported-by-changed"
        if forward:
            python_appendum = "transitive import by changed Python file"
        else:
            python_appendum = "transitively imports changed Python file"
        add_python_deps(changes_ex, forward=forward, cwd=cwd)
        python_deps = set(changes_ex) - original

    print(f"Changes with respect to {base_branch} branch:")
    if changes_ex:
        for file in sorted(changes_ex):
            appendums = []
            if file in python_deps:
                appendums.append(python_appendum)
            appendum = f" ({', '.join(appendums)})" if appendums else ""
            print(f" * {file}{appendum}")
    else:
        print("  (no changes found)")
    return 0


# Mapping from command line subcommand names to functions which implement those.
# The functions need to take a single argument, the processed list of args.
ARGS_MAP: dict[str, Callable[[argparse.Namespace], int]] = {
    "lint-config": lint_config,
    "init": create_initial_config,
    "show-changes": show_changes,
}


class InvalidArgumentError(Exception):
    """
    Error while parsing arguments.
    """


def parse_args(program_name: str, args: list[str]) -> argparse.Namespace:
    """
    Parse the command line arguments.
    """

    toplevel_parser = argparse.ArgumentParser(
        prog=program_name,
        description="Script to manage generated documentation for ansible",
    )
    toplevel_parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the antsibull-nox version",
    )
    subparsers = toplevel_parser.add_subparsers(
        title="Subcommands", dest="command", help="for help use: `SUBCOMMANDS -h`"
    )
    subparsers.required = True

    subparsers.add_parser(
        "lint-config",
        description="Lint antsibull-nox configuration file",
    )

    subparsers.add_parser(
        "init",
        description="Create noxfile and antsibull-nox configuration file",
    )

    show_changes_parser = subparsers.add_parser(
        "show-changes", description="Show changed files"
    )
    show_changes_parser.add_argument(
        "--add-python-deps",
        default="none",
        choices=["none", "imported-by-changed", "importing-changed"],
        help="Include Python dependencies (forward or backward) for changed Python files."
        " If set to 'imported-by-changed', all Python files (transitively) importing"
        " changed files will be added ('backward'). If set to 'importing-changed', all"
        " Python files (transitively) imported by changed files will be added ('forward').",
    )

    # This must come after all parser setup
    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(toplevel_parser)

    parsed_args: argparse.Namespace = toplevel_parser.parse_args(args)
    return parsed_args


def run(args: list[str]) -> int:
    """
    Run the program.
    """
    program_name = os.path.basename(args[0])
    try:
        parsed_args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e, file=sys.stderr)
        return 2

    return ARGS_MAP[parsed_args.command](parsed_args)


def main() -> int:
    """
    Entrypoint called from the script.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
    """
    return run(sys.argv)


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

from __future__ import annotations

import typing as t

import pytest

from antsibull_nox.messages import (
    Level,
    Location,
    Message,
)
from antsibull_nox.sessions.utils.output import (
    _BOLD,
    _FAINT,
    _ITALICS,
    _RESET,
    _UNDERLINE,
    ASCII_MARKINGS,
    Color,
    ColorComposer,
    Markings,
    _compose_first_line,
    _compose_message_with_note,
    _determine_marker,
    _Formatting,
    _render_code,
    format_messages_plain,
    should_fail,
    split_lines,
    split_lines_with_prefix,
)

SPLIT_LINES_DATA: list[tuple[str, list[str]]] = [
    (
        "",
        [],
    ),
    (
        "\n\n",
        [],
    ),
    (
        "Foo",
        ["Foo"],
    ),
    (
        "Foo\n",
        ["Foo"],
    ),
    (
        "\nFoo",
        ["", "Foo"],
    ),
    (
        "Foo\r\nBar",
        ["Foo", "Bar"],
    ),
    (
        "Foo\rBar",
        ["Foo", "Bar"],
    ),
    (
        "Foo\n\rBar",
        ["Foo", "", "Bar"],
    ),
]


@pytest.mark.parametrize(
    "text, expected_result",
    SPLIT_LINES_DATA,
)
def test_split_lines(text: str, expected_result: list[str]) -> None:
    assert split_lines(text) == expected_result


SPLIT_LINES_WITH_PREFIX_DATA: list[tuple[str, dict[str, t.Any], list[str]]] = [
    (
        "",
        {},
        [],
    ),
    (
        "\n\n",
        {},
        [],
    ),
    (
        "",
        {"at_least_one_line": True},
        [""],
    ),
    (
        "\n\n",
        {"at_least_one_line": True},
        [""],
    ),
    (
        "Foo",
        {},
        [" Foo"],
    ),
    (
        "Foo",
        {"separator": ""},
        ["Foo"],
    ),
    (
        "Foo\nBar",
        {"prefix": "A", "separator": "B"},
        ["ABFoo", " BBar"],
    ),
    (
        "Foo\nBar\n\nBaz\n\n\n",
        {"prefix": "Abc"},
        ["Abc Foo", "    Bar", "    ", "    Baz"],
    ),
]


@pytest.mark.parametrize(
    "text, kwargs, expected_result",
    SPLIT_LINES_WITH_PREFIX_DATA,
)
def test_split_lines_with_prefix(
    text: str, kwargs: dict[str, t.Any], expected_result: list[str]
) -> None:
    assert list(split_lines_with_prefix(text, **kwargs)) == expected_result


SHOULD_FAIL_DATA: list[tuple[list[Message], bool]] = [
    ([], False),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
        ],
        False,
    ),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.WARNING,
                id=None,
                message="",
            ),
        ],
        True,
    ),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="",
            ),
        ],
        True,
    ),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
        ],
        False,
    ),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.WARNING,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
        ],
        True,
    ),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="",
            ),
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
            ),
        ],
        True,
    ),
]


@pytest.mark.parametrize(
    "messages, expected_result",
    SHOULD_FAIL_DATA,
)
def test_should_fail(messages: list[Message], expected_result: bool) -> None:
    assert should_fail(messages) == expected_result


FORMAT_MESSAGES_PLAIN_DATA: list[tuple[list[Message], list[str]]] = [
    ([], []),
    (
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        [":0:0:"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:0:0:"],
    ),
    (
        [
            Message(
                file=None,
                position=Location(line=5),
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        [":5:0:"],
    ),
    (
        [
            Message(
                file="foo",
                position=Location(line=5),
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:5:0:"],
    ),
    (
        [
            Message(
                file="foo",
                position=Location(line=5, column=10),
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:5:10:"],
    ),
    (
        [
            Message(
                file="foo",
                position=Location(line=5, exact=False),
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:~5:0:"],
    ),
    (
        [
            Message(
                file="foo",
                position=Location(line=5, column=10, exact=False),
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:~5:10:"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id="bar",
                message="",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:0:0: [bar]"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="bar",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:0:0: bar"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol="bar",
                hint=None,
                note=None,
                url=None,
            ),
        ],
        ["foo:0:0:  [bar]"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint="bar",
                note=None,
                url=None,
            ),
        ],
        ["foo:0:0: ", "         bar"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note="bar",
                url=None,
            ),
        ],
        ["foo:0:0: ", "         Note: bar"],
    ),
    (
        [
            Message(
                file="foo",
                position=None,
                end_position=None,
                level=Level.INFO,
                id=None,
                message="",
                symbol=None,
                hint=None,
                note=None,
                url="bar",
            ),
        ],
        ["foo:0:0:"],
    ),
    (
        [
            Message(
                file="foo",
                position=Location(line=1, column=2),
                end_position=Location(line=3, column=4),
                level=Level.ERROR,
                id="foo",
                message="bar",
                symbol="baz",
                hint="bam",
                note="boom",
                url="fubar",
            ),
        ],
        ["foo:1:2: [foo] bar [baz]", "               bam", "               Note: boom"],
    ),
]


@pytest.mark.parametrize(
    "messages, expected_result",
    FORMAT_MESSAGES_PLAIN_DATA,
)
def test_format_messages_plain(
    messages: list[Message], expected_result: list[str]
) -> None:
    assert list(format_messages_plain(messages)) == expected_result


def test_formatting() -> None:
    assert not _Formatting(color=Color.RED).eq_color(_Formatting(color=Color.GREEN))
    assert _Formatting(bold=True).eq_color(_Formatting(bold=False))
    assert _Formatting(faint=True).eq_color(_Formatting(faint=False))
    assert _Formatting(italics=True).eq_color(_Formatting(italics=False))
    assert _Formatting(underline=True).eq_color(_Formatting(underline=False))

    assert _Formatting(color=Color.RED).eq_formatting(_Formatting(color=Color.GREEN))
    assert not _Formatting(bold=True).eq_formatting(_Formatting(bold=False))
    assert not _Formatting(faint=True).eq_formatting(_Formatting(faint=False))
    assert not _Formatting(italics=True).eq_formatting(_Formatting(italics=False))
    assert not _Formatting(underline=True).eq_formatting(_Formatting(underline=False))

    assert _Formatting().update(color=Color.RED) == _Formatting(color=Color.RED)
    assert _Formatting().update(bold=True) == _Formatting(bold=True)
    assert _Formatting().update(faint=True) == _Formatting(faint=True)
    assert _Formatting(faint=True).update(bold=True) == _Formatting(bold=True)
    assert _Formatting(bold=True).update(faint=True) == _Formatting(faint=True)
    assert _Formatting().update(italics=True) == _Formatting(italics=True)
    assert _Formatting().update(underline=True) == _Formatting(underline=True)

    with pytest.raises(ValueError):
        _Formatting().update(bold=True, faint=True)

    assert _Formatting(color=Color.RED).format(use_color=True, use_formatting=True) == [
        Color.RED.value
    ]
    assert _Formatting(bold=True).format(use_color=True, use_formatting=True) == [_BOLD]
    assert _Formatting(faint=True).format(use_color=True, use_formatting=True) == [
        _FAINT
    ]
    assert _Formatting(italics=True).format(use_color=True, use_formatting=True) == [
        _ITALICS
    ]
    assert _Formatting(underline=True).format(use_color=True, use_formatting=True) == [
        _UNDERLINE
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(use_color=True, use_formatting=True) == [
        _BOLD,
        _FAINT,
        _ITALICS,
        _UNDERLINE,
        Color.RED.value,
    ]

    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(color=Color.RED)
    ) == [
        _BOLD,
        _FAINT,
        _ITALICS,
        _UNDERLINE,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(use_color=True, use_formatting=True, previous=_Formatting(bold=True)) == [
        _FAINT,
        _ITALICS,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(use_color=True, use_formatting=True, previous=_Formatting(faint=True)) == [
        _BOLD,
        _ITALICS,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(italics=True)
    ) == [
        _BOLD,
        _FAINT,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=True
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(underline=True)
    ) == [
        _BOLD,
        _FAINT,
        _ITALICS,
        Color.RED.value,
    ]

    assert _Formatting(
        color=Color.NONE, bold=True, faint=True, italics=True, underline=True
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(color=Color.RED)
    ) == [
        _RESET,
        _BOLD,
        _FAINT,
        _ITALICS,
        _UNDERLINE,
    ]
    assert _Formatting(
        color=Color.RED, bold=False, faint=True, italics=True, underline=True
    ).format(use_color=True, use_formatting=True, previous=_Formatting(bold=True)) == [
        _RESET,
        _FAINT,
        _ITALICS,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=False, italics=True, underline=True
    ).format(use_color=True, use_formatting=True, previous=_Formatting(faint=True)) == [
        _RESET,
        _BOLD,
        _ITALICS,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=False, underline=True
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(italics=True)
    ) == [
        _RESET,
        _BOLD,
        _FAINT,
        _UNDERLINE,
        Color.RED.value,
    ]
    assert _Formatting(
        color=Color.RED, bold=True, faint=True, italics=True, underline=False
    ).format(
        use_color=True, use_formatting=True, previous=_Formatting(underline=True)
    ) == [
        _RESET,
        _BOLD,
        _FAINT,
        _ITALICS,
        Color.RED.value,
    ]


def test_color_composer() -> None:
    # empty
    assert (
        ColorComposer(use_color=True, use_formatting=True).add_text("", bold=True).empty
    )
    assert (
        not ColorComposer(use_color=True, use_formatting=True)
        .add_text("foo", bold=True)
        .empty
    )

    assert (
        ColorComposer(use_color=True, use_formatting=True)
        .add_text("foo", bold=True)
        .add_text("bar", faint=True)
        .value
        == f"{_BOLD}foo{_RESET}{_FAINT}bar{_RESET}"
    )
    assert (
        ColorComposer(use_color=True, use_formatting=True)
        .add_text("foo", bold=True)
        .add_text("bar", color=Color.RED)
        .value
        == f"{_BOLD}foo{Color.RED.value}bar{_RESET}"
    )


DETERMINE_MARKER_DATA: list[
    tuple[
        bool,
        int | None,
        int,
        tuple[int, int | None],
        tuple[int, int | None] | None,
        str,
        Markings,
        bool,
        bool,
        str,
        str,
        str,
    ]
] = [
    (
        False,
        None,
        0,
        (0, None),
        None,
        "  ",
        ASCII_MARKINGS,
        False,
        False,
        "",
        "",
        "  ",
    ),
    (
        True,
        None,
        0,
        (0, None),
        None,
        "  ",
        ASCII_MARKINGS,
        True,
        False,
        "",
        "",
        "| ",
    ),
    (
        False,
        123,
        23,
        (122, None),
        None,
        "  ",
        ASCII_MARKINGS,
        False,
        False,
        "",
        "",
        "  ",
    ),
    (
        False,
        123,
        23,
        (123, None),
        None,
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "",
        "",
        "> ",
    ),
    (
        False,
        123,
        23,
        (123, 3),
        None,
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "",
        "    ^",
        "  ",
    ),
    (
        False,
        123,
        23,
        (123, None),
        (123, None),
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "/",
        "\\",
        "| ",
    ),
    (
        False,
        123,
        23,
        (123, 2),
        (123, 4),
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "",
        "   ^^^",
        "  ",
    ),
    (
        False,
        123,
        23,
        (123, 2),
        (124, 4),
        "  ",
        ASCII_MARKINGS,
        True,
        False,
        "/",
        "",
        "| ",
    ),
    (
        True,
        124,
        23,
        (123, 2),
        (124, 4),
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "",
        "\\-----/",
        "| ",
    ),
    (
        False,
        123,
        23,
        (123, 2),
        (124, None),
        "  ",
        ASCII_MARKINGS,
        True,
        False,
        "/",
        "",
        "| ",
    ),
    (
        True,
        124,
        23,
        (123, 2),
        (124, None),
        "  ",
        ASCII_MARKINGS,
        True,
        True,
        "",
        "\\",
        "| ",
    ),
]


@pytest.mark.parametrize(
    "inside, line_no, line_length, content_start, content_end, extra_indent, markings,"
    " expected_inside, expected_last_inside, expected_before_mark, expected_after_mark,"
    " expected_this_extra_indent",
    DETERMINE_MARKER_DATA,
)
def test__determine_marker(
    inside: bool,
    line_no: int | None,
    line_length: int,
    content_start: tuple[int, int | None],
    content_end: tuple[int, int | None] | None,
    extra_indent: str,
    markings: Markings,
    expected_inside: bool,
    expected_last_inside: bool,
    expected_before_mark: str,
    expected_after_mark: str,
    expected_this_extra_indent: str,
) -> None:
    result = _determine_marker(
        inside, line_no, line_length, content_start, content_end, extra_indent, markings
    )
    assert result == (
        expected_inside,
        expected_last_inside,
        expected_before_mark,
        expected_after_mark,
        expected_this_extra_indent,
    )


RENDER_CODE_DATA: list[
    tuple[
        list[tuple[int | None, str]],
        tuple[int, int | None],
        tuple[int, int | None] | None,
        Markings,
        list[str],
    ]
] = [
    (
        [(None, "foo")],
        (0, None),
        None,
        ASCII_MARKINGS,
        [": |   foo"],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, None),
        None,
        ASCII_MARKINGS,
        [
            ": |   ",
            "3 |   foo",
            "4 | > bar",
            "5 |   baz",
            "6 |   bam",
            ": |   ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, None),
        (4, None),
        ASCII_MARKINGS,
        [
            ": |   ",
            "3 |   foo",
            "  | /",
            "4 | | bar",
            "  | \\",
            "5 |   baz",
            "6 |   bam",
            ": |   ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, None),
        (5, None),
        ASCII_MARKINGS,
        [
            ": |   ",
            "3 |   foo",
            "  | /",
            "4 | | bar",
            "5 | | baz",
            "  | \\",
            "6 |   bam",
            ": |   ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, 2),
        None,
        ASCII_MARKINGS,
        [
            ": | ",
            "3 | foo",
            "4 | bar",
            "  |  ^",
            "5 | baz",
            "6 | bam",
            ": | ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "barbam"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, 2),
        (4, 3),
        ASCII_MARKINGS,
        [
            ": | ",
            "3 | foo",
            "4 | barbam",
            "  |  ^^",
            "5 | baz",
            "6 | bam",
            ": | ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "barbam"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, 4),
        (4, -1),
        ASCII_MARKINGS,
        [
            ": | ",
            "3 | foo",
            "4 | barbam",
            "  |    ^^^^",
            "5 | baz",
            "6 | bam",
            ": | ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, 3),
        (5, 2),
        ASCII_MARKINGS,
        [
            ": |   ",
            "3 |   foo",
            "  | /",
            "4 | | bar",
            "5 | | baz",
            "  | \\---/",
            "6 |   bam",
            ": |   ",
        ],
    ),
    (
        [
            (None, ""),
            (3, "foo"),
            (4, "bar"),
            (5, "baz"),
            (6, "bam"),
            (None, ""),
        ],
        (4, 3),
        (5, -1),
        ASCII_MARKINGS,
        [
            ": |   ",
            "3 |   foo",
            "  | /",
            "4 | | bar",
            "5 | | baz",
            "  | \\-----/",
            "6 |   bam",
            ": |   ",
        ],
    ),
]


@pytest.mark.parametrize(
    "content, content_start, content_end, markings, expected_result",
    RENDER_CODE_DATA,
)
def test__render_code(
    content: list[tuple[int | None, str]],
    content_start: tuple[int, int | None],
    content_end: tuple[int, int | None] | None,
    markings: Markings,
    expected_result: list[str],
) -> None:
    result = list(
        _render_code(
            content,
            content_start,
            content_end,
            lambda: ColorComposer(use_color=False, use_formatting=False),
            markings,
        )
    )
    assert result == expected_result


COMPOSE_FIRST_LINE_DATA: list[tuple[Message, str]] = [
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
        ),
        "",
    ),
    (
        Message(
            file=None,
            position=Location(line=1, column=2),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id=None,
            message="",
        ),
        "",
    ),
    (
        Message(
            file="foo",
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
        ),
        "foo:",
    ),
    (
        Message(
            file="foo",
            position=Location(line=1, column=None),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id=None,
            message="",
        ),
        "foo:1:",
    ),
    (
        Message(
            file="foo",
            position=Location(line=1, column=None, exact=False),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id=None,
            message="",
        ),
        "foo:~1:",
    ),
    (
        Message(
            file="foo",
            position=Location(line=1, column=2),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id=None,
            message="",
        ),
        "foo:1:2:",
    ),
    (
        Message(
            file="foo",
            position=Location(line=1, column=2, exact=False),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id=None,
            message="",
        ),
        "foo:~1:2:",
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id="bar",
            message="",
        ),
        "[bar]",
    ),
    (
        Message(
            file="foo",
            position=None,
            end_position=None,
            level=Level.INFO,
            id="bar",
            message="",
        ),
        "foo: [bar]",
    ),
    (
        Message(
            file="foo",
            position=Location(line=1, column=2),
            end_position=Location(line=3, column=4),
            level=Level.INFO,
            id="bar",
            message="",
        ),
        "foo:1:2: [bar]",
    ),
]


@pytest.mark.parametrize(
    "message, expected_result",
    COMPOSE_FIRST_LINE_DATA,
)
def test__compose_first_line(
    message: Message,
    expected_result: str,
) -> None:
    result = _compose_first_line(
        message, lambda: ColorComposer(use_color=False, use_formatting=False)
    ).value
    assert result == expected_result


COMPOSE_MESSAGE_WITH_NOTE_DATA: list[tuple[Message, str, bool, bool, list[str]]] = [
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint=None,
            note=None,
            url=None,
        ),
        "<>",
        False,
        False,
        [],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint=None,
            note=None,
            url=None,
        ),
        "<>",
        False,
        True,
        ["<>"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="foo",
            symbol=None,
            hint=None,
            note=None,
            url=None,
        ),
        "<>",
        False,
        False,
        ["<>foo"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint=None,
            note="foo",
            url=None,
        ),
        "<>",
        False,
        True,
        ["<>Note: foo"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint=None,
            note="foo\nbar",
            url=None,
        ),
        "<>",
        False,
        True,
        ["<>Note: foo", "<>bar"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint="foo",
            note=None,
            url=None,
        ),
        "<>",
        False,
        False,
        [],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="",
            symbol=None,
            hint="foo",
            note=None,
            url=None,
        ),
        "<>",
        False,
        True,
        ["<>"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="foo\nbar",
            symbol="bam",
            hint="baz",
            note="boom",
            url="bang",
        ),
        "<>",
        False,
        False,
        ["<>foo", "<>bar [bam]", "<>Note: boom"],
    ),
    (
        Message(
            file=None,
            position=None,
            end_position=None,
            level=Level.INFO,
            id=None,
            message="foo\nbar",
            symbol="bam",
            hint="baz",
            note="boom",
            url="bang",
        ),
        "<>",
        True,
        False,
        ["<>foo", "<>bar [bam]", "<>baz", "<>Note: boom"],
    ),
]


@pytest.mark.parametrize(
    "message, indent, add_hint, at_least_one_line, expected_result",
    COMPOSE_MESSAGE_WITH_NOTE_DATA,
)
def test__compose_message_with_note(
    message: Message,
    indent: str,
    add_hint: bool,
    at_least_one_line: bool,
    expected_result: list[str],
) -> None:
    result = list(
        _compose_message_with_note(
            message,
            indent,
            lambda: ColorComposer(use_color=False, use_formatting=False),
            add_hint=add_hint,
            at_least_one_line=at_least_one_line,
        )
    )
    assert result == expected_result



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

from __future__ import annotations

from pathlib import Path

import pytest

from antsibull_nox.messages import Level, Location, Message
from antsibull_nox.messages.parse import (
    parse_antsibull_docs_errors,
    parse_bare_framework_errors,
    parse_mypy_errors,
    parse_pylint_json2_errors,
    parse_ruff_check_errors,
)

PARSE_ANTSIBULL_DOCS_ERRORS_DATA: list[tuple[str, list[Message]]] = [
    (
        r"""{"messages": []}""",
        [],
    ),
    (
        r"""
Some random text.
{
  "messages": [
    {
      "path": "docs/docsite/config.yml",
      "row": null,
      "column": null,
      "message": "changelog -> write_changelog: Input should be a valid boolean, unable to interpret input"
    },
    {
      "path": "docs/docsite/extra-docs.yml",
      "row": null,
      "column": null,
      "message": "Section #0 has no content"
    },
    {
      "path": "docs/docsite/extra-docs.yml",
      "row": null,
      "column": null,
      "message": "Toctree entry in section #0 is not a list"
    },
    {
      "path": "docs/docsite/links.yml",
      "row": null,
      "column": null,
      "message": "foo: Extra inputs are not permitted"
    },
    {
      "path": "docs/docsite/rst/filter_guide.rst",
      "row": 6,
      "column": null,
      "message": "Label \"ansible_collection.community.dns.docsite.filter_guide\" does not start with expected prefix \"ansible_collections.community.dns.docsite.\""
    },
    {
      "path": "plugins/modules/adguardhome_rewrite.py",
      "row": null,
      "column": null,
      "message": "DOCUMENTATION -> options -> state -> description[2]: O(foo): option name does not reference to an existing option of the module community.dns.adguardhome_rewrite"
    },
    {
      "path": "foo/bar",
      "row": 2,
      "column": 3,
      "end_column": null,
      "message": "bar"
    },
    {
      "path": "foo/bar",
      "row": 5,
      "column": 6,
      "end_column": 10,
      "message": "baz"
    },
    {
      "path": "foo/bar",
      "row": 10,
      "column": null,
      "end_column": 15,
      "message": "bam"
    },
    {
      "path": "foo/bar",
      "row": 12,
      "column": null,
      "end_column": null,
      "message": "bino"
    }
  ],
  "success": false
}
More random text.
""",
        [
            Message(
                file="docs/docsite/config.yml",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=(
                    "changelog -> write_changelog: Input should be a valid"
                    " boolean, unable to interpret input"
                ),
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="docs/docsite/extra-docs.yml",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Section #0 has no content",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="docs/docsite/extra-docs.yml",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Toctree entry in section #0 is not a list",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="docs/docsite/links.yml",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="foo: Extra inputs are not permitted",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="docs/docsite/rst/filter_guide.rst",
                position=Location(line=6, column=None),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=(
                    'Label "ansible_collection.community.dns.docsite.filter_guide"'
                    " does not start with expected prefix"
                    ' "ansible_collections.community.dns.docsite."'
                ),
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="plugins/modules/adguardhome_rewrite.py",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message=(
                    "DOCUMENTATION -> options -> state -> description[2]: O(foo):"
                    " option name does not reference to an existing option of the"
                    " module community.dns.adguardhome_rewrite"
                ),
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="foo/bar",
                position=Location(line=2, column=3),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="bar",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="foo/bar",
                position=Location(line=5, column=6),
                end_position=Location(line=5, column=10),
                level=Level.ERROR,
                id=None,
                message="baz",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="foo/bar",
                position=Location(line=10),
                end_position=Location(line=10, column=15),
                level=Level.ERROR,
                id=None,
                message="bam",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="foo/bar",
                position=Location(line=12),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="bino",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""Bad output.""",
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Cannot parse antsibull-docs lint-collection-docs output: "
                "Expecting value: line 1 column 1 (char 0)\n"
                "Bad output.",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    PARSE_ANTSIBULL_DOCS_ERRORS_DATA,
)
def test_parse_antsibull_docs_errors(
    output: str, expected_result: list[Message]
) -> None:
    result = parse_antsibull_docs_errors(output=output)
    print(result)
    assert result == expected_result


PARSE_BARE_FRAMEWORK_ERRORS_DATA: list[tuple[str, list[Message]]] = [
    (
        r"""{"messages": []}""",
        [],
    ),
    (
        r"""{"messages": [{"file": "meta/runtime.yml", "start": null, "end": null, "level": "error", "id": null,"""
        r""" "message": "module 'hosttech_dns_record' is not part of 'hosttech' action group", "symbol": null,"""
        r""" "hint": null, "note": null, "url": null}]}""",
        [
            Message(
                file="meta/runtime.yml",
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="module 'hosttech_dns_record' is not part of 'hosttech' action "
                "group",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""{"messages": [{"file": "plugins/inventory/hosttech_dns_records.py", "start": {"line": 16, "column": null},"""
        r""" "end": null, "level": "error", "id": null, "message": "found trailing whitespace", "symbol": null, "hint":"""
        r""" null, "note": null, "url": null}]}""",
        [
            Message(
                file="plugins/inventory/hosttech_dns_records.py",
                position=Location(
                    line=16,
                    column=None,
                ),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="found trailing whitespace",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""Bad output.""",
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Cannot parse output: Expecting value: line 1 column 1 (char "
                "0)\n"
                "Bad output.",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""{"messages": [{"file": "docs/docsite/rst/hosttech_guide.rst", "start": {"line": 113, "column": 1, "exact": false},"""
        r""""end": null, "level": "error", "id": null, "message": "error: syntax error: could not find expected ':' (syntax)","""
        r""""symbol": null, "hint": null, "note": "The code block could not be exactly located in the source file. """
        r"""The line/column numbers might be off.", "url": null}, {"file": "docs/docsite/rst/hosttech_guide.rst", "start":"""
        r""" {"line": 123, "column": 1, "exact": true}, "end": null, "level": "error", "id": null, "message": "error: syntax"""
        r""" error: could not find expected ':' (syntax)", "symbol": null, "hint": null, "note": null, "url": null}]}""",
        [
            Message(
                file="docs/docsite/rst/hosttech_guide.rst",
                position=Location(line=113, column=1, exact=False),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="error: syntax error: could not find expected ':' (syntax)",
                symbol=None,
                hint=None,
                note="The code block could not be exactly located in the source file. The line/column numbers might be off.",
                url=None,
            ),
            Message(
                file="docs/docsite/rst/hosttech_guide.rst",
                position=Location(line=123, column=1, exact=True),
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="error: syntax error: could not find expected ':' (syntax)",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    PARSE_BARE_FRAMEWORK_ERRORS_DATA,
)
def test_parse_bare_framework_errors(
    output: str, expected_result: list[Message]
) -> None:
    result = parse_bare_framework_errors(output=output)
    print(result)
    assert result == expected_result


PARSE_MYPY_ERRORS_DATA: list[tuple[str, list[Message]]] = [
    (
        r"""""",
        [],
    ),
    (
        r"""
""",
        [],
    ),
    (
        r"""{"file": "ansible_collections/community/general/plugins/cache/memcached.py", "line": 254,"""
        r""" "column": 11, "message": "Incompatible types in assignment (expression has type \"None\","""
        r""" variable has type \"str\")", "hint": null, "code": "assignment", "severity": "error"}""",
        [
            Message(
                file="plugins/cache/memcached.py",
                position=Location(line=254, column=12),
                end_position=None,
                level=Level.ERROR,
                id="assignment",
                message='Incompatible types in assignment (expression has type "None", variable has type "str")',
                symbol=None,
                hint=None,
                note=None,
                url=None,
            )
        ],
    ),
    (
        r"""Bad output.""",
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Cannot parse mypy output: Bad output.",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    PARSE_MYPY_ERRORS_DATA,
)
def test_parse_mypy_errors(output: str, expected_result: list[Message]) -> None:
    result = parse_mypy_errors(
        output=output,
        root_path=Path("/root"),
        source_path=Path("/root/ansible_collections/community/general"),
    )
    print(result)
    assert result == expected_result


PARSE_PYLINT_JSON2_ERRORS_DATA: list[tuple[str, list[Message]]] = [
    (
        r"""{"messages": []}""",
        [],
    ),
    (
        r"""
{
    "messages": [
        {
            "type": "warning",
            "symbol": "pointless-statement",
            "message": "Statement seems to have no effect",
            "messageId": "W0104",
            "confidence": "UNDEFINED",
            "module": "ansible_collections.community.dns.plugins.module_utils.ips",
            "obj": "",
            "line": 43,
            "column": 0,
            "endLine": 43,
            "endColumn": 4,
            "path": "ansible_collections/community/dns/plugins/module_utils/ips.py",
            "absolutePath": "/root/ansible_collections/community/dns/plugins/module_utils/ips.py"
        },
        {
            "type": "warning",
            "symbol": "pointless-statement",
            "message": "Statement seems to have no effect II",
            "messageId": "W0104",
            "confidence": "UNDEFINED",
            "module": "ansible_collections.community.dns.plugins.module_utils.ips",
            "obj": "",
            "line": 47,
            "column": 3,
            "path": "ansible_collections/community/dns/plugins/module_utils/ips.py",
            "absolutePath": "/root/ansible_collections/community/dns/plugins/module_utils/ips.py"
        }
    ],
    "statistics": {
        "messageTypeCount": {
            "fatal": 0,
            "error": 0,
            "warning": 1,
            "refactor": 0,
            "convention": 0,
            "info": 0
        },
        "modulesLinted": 92,
        "score": 9.998451772720236
    }
}
""",
        [
            Message(
                file="plugins/module_utils/ips.py",
                position=Location(
                    line=43,
                    column=0,
                ),
                end_position=Location(
                    line=43,
                    column=4,
                ),
                level=Level.ERROR,
                id="W0104",
                message="Statement seems to have no effect",
                symbol="pointless-statement",
                hint=None,
                note=None,
                url=None,
            ),
            Message(
                file="plugins/module_utils/ips.py",
                position=Location(
                    line=47,
                    column=3,
                ),
                end_position=None,
                level=Level.ERROR,
                id="W0104",
                message="Statement seems to have no effect II",
                symbol="pointless-statement",
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""Bad output.""",
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Cannot parse pylint output: Expecting value: line 1 column 1 "
                "(char 0)\n"
                "Bad output.",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    PARSE_PYLINT_JSON2_ERRORS_DATA,
)
def test_parse_pylint_json2_errors(output: str, expected_result: list[Message]) -> None:
    result = parse_pylint_json2_errors(
        output=output, source_path=Path("/root/ansible_collections/community/dns")
    )
    print(result)
    assert result == expected_result


PARSE_RUFF_CHECK_ERRORS_DATA: list[tuple[str, list[Message]]] = [
    (
        r"""[]""",
        [],
    ),
    (
        r"""[
  {
    "cell": null,
    "code": "SIM222",
    "end_location": {
      "column": 28,
      "row": 245
    },
    "filename": "/ansible_collections/community/general/plugins/cache/memcached.py",
    "fix": {
      "applicability": "unsafe",
      "edits": [
        {
          "content": "True",
          "end_location": {
            "column": 28,
            "row": 245
          },
          "location": {
            "column": 16,
            "row": 245
          }
        }
      ],
      "message": "Replace with `True`"
    },
    "location": {
      "column": 16,
      "row": 245
    },
    "message": "Use `True` instead of `True or ...`",
    "noqa_row": 245,
    "url": "https://docs.astral.sh/ruff/rules/expr-or-true"
  },
  {
    "cell": null,
    "code": "SIM103",
    "end_location": {
      "column": 25,
      "row": 251
    },
    "filename": "/ansible_collections/community/general/plugins/cache/memcached.py",
    "fix": {
      "applicability": "unsafe",
      "edits": [
        {
          "content": "return bool(foo)",
          "end_location": {
            "column": 25,
            "row": 251
          },
          "location": {
            "column": 9,
            "row": 248
          }
        }
      ],
      "message": "Replace with `return bool(foo)`"
    },
    "location": {
      "column": 9,
      "row": 248
    },
    "message": "Return the condition `bool(foo)` directly",
    "noqa_row": 248,
    "url": "https://docs.astral.sh/ruff/rules/needless-bool"
  }
]
""",
        [
            Message(
                file="plugins/cache/memcached.py",
                position=Location(line=245, column=16),
                end_position=Location(line=245, column=27),
                level=Level.ERROR,
                id="SIM222",
                message="Use `True` instead of `True or ...`",
                symbol=None,
                hint="Replace with `True`",
                note=None,
                url="https://docs.astral.sh/ruff/rules/expr-or-true",
            ),
            Message(
                file="plugins/cache/memcached.py",
                position=Location(line=248, column=9),
                end_position=Location(line=251, column=24),
                level=Level.ERROR,
                id="SIM103",
                message="Return the condition `bool(foo)` directly",
                symbol=None,
                hint="Replace with `return bool(foo)`",
                note=None,
                url="https://docs.astral.sh/ruff/rules/needless-bool",
            ),
        ],
    ),
    (
        r"""
[
  {
    "cell": null,
    "code": "invalid-syntax",
    "end_location": {
      "column": 1,
      "row": 243
    },
    "filename": "/ansible_collections/community/general/plugins/cache/memcached.py",
    "fix": null,
    "location": {
      "column": 33,
      "row": 242
    },
    "message": "Expected `:`, found newline",
    "noqa_row": null,
    "url": null
  }
]
""",
        [
            Message(
                file="plugins/cache/memcached.py",
                position=Location(
                    line=242,
                    column=33,
                ),
                end_position=Location(
                    line=242,
                    column=-1,
                ),
                level=Level.ERROR,
                id="invalid-syntax",
                message="Expected `:`, found newline",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
    (
        r"""Bad output.""",
        [
            Message(
                file=None,
                position=None,
                end_position=None,
                level=Level.ERROR,
                id=None,
                message="Cannot parse ruff check output: Expecting value: line 1 "
                "column 1 (char 0)\n"
                "Bad output.",
                symbol=None,
                hint=None,
                note=None,
                url=None,
            ),
        ],
    ),
]


@pytest.mark.parametrize(
    "output, expected_result",
    PARSE_RUFF_CHECK_ERRORS_DATA,
)
def test_parse_ruff_check_errors(output: str, expected_result: list[Message]) -> None:
    result = parse_ruff_check_errors(
        output=output, source_path=Path("/ansible_collections/community/general")
    )
    print(result)
    assert result == expected_result



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

"""
Tests for the collection module
"""

from __future__ import annotations

import typing as t
from pathlib import Path

from antsibull_fileutils.yaml import store_yaml_file

from antsibull_nox.collection.install import (
    Runner,
)


def create_collection(
    path: Path,
    *,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> None:
    data: dict[str, t.Any] = {
        "namespace": namespace,
        "name": name,
    }
    if version is not None:
        data["version"] = version
    if dependencies is not None:
        data["dependencies"] = dependencies
    path.mkdir(parents=True, exist_ok=True)
    store_yaml_file(path / "galaxy.yml", data)


def create_collection_w_dir(
    root: Path,
    *,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> Path:
    path = root / namespace / name
    create_collection(
        path=path,
        namespace=namespace,
        name=name,
        version=version,
        dependencies=dependencies,
    )
    return path


def create_collection_w_shallow_dir(
    root: Path,
    *,
    directory_override: str | None = None,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> Path:
    path = root / (directory_override or f"{namespace}.{name}")
    create_collection(
        path=path,
        namespace=namespace,
        name=name,
        version=version,
        dependencies=dependencies,
    )
    return path


def create_once_runner(
    expected_args: list[str],
    stdout: bytes,
    stderr: bytes = b"",
    expect_check: bool = True,
    expect_use_venv_if_present: bool = True,
    rc: int = 0,
) -> Runner:
    call_counter = [0]

    def runner(
        args: list[str], *, check: bool = True, use_venv_if_present: bool = True
    ) -> tuple[bytes, bytes, int]:
        assert call_counter[0] == 0
        assert args == expected_args
        assert check is expect_check
        assert use_venv_if_present is expect_use_venv_if_present
        call_counter[0] += 1
        return stdout, stderr, rc

    return runner



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
VCS interfaces and types.
"""

from __future__ import annotations

import abc
import typing as t
from pathlib import Path


class VcsProvider(metaclass=abc.ABCMeta):
    """
    Provides an abstract interface to VCSes.
    """

    @abc.abstractmethod
    def find_repo_path(self, *, path: Path) -> Path | None:
        """
        Given a path, finds the associated repository path for this VCS,
        if there is one.
        """

    @abc.abstractmethod
    def get_current_branch(self, *, repo: Path) -> str | None:
        """
        Retrieve the current branch of the repository.

        If there is no current branch for some reason, ``None`` is returned.
        """

    @abc.abstractmethod
    def get_changes_compared_to(self, *, repo: Path, branch: str) -> list[Path]:
        """
        Retrieve a list of files and directories that have been added, changed, or removed.

        The returned paths are relative to ``repo``!
        """


VCS = t.Literal["git"]


__all__ = (
    "VcsProvider",
    "VCS",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox docs check session.
"""

from __future__ import annotations

import typing as t
from pathlib import Path

import nox

from ..cd import (
    get_changes,
)
from ..config import CONFIG_FILENAME
from ..messages.parse import parse_antsibull_docs_errors
from ..paths.utils import (
    list_all_files,
)
from .collections import (
    CollectionSetup,
    prepare_collections,
)
from .utils import silence_run_verbosity
from .utils.output import print_messages
from .utils.package_versions import (
    get_package_version,
    is_new_enough,
)
from .utils.packages import (
    PackageType,
    PackageTypeOrList,
    check_package_types,
    install,
    normalize_package_type,
)
from .utils.scripts import (
    run_bare_script,
)


def find_extra_docs_rst_files() -> list[Path]:
    """
    Find all RST extra document files.
    """
    all_files = list_all_files()
    cwd = Path.cwd()
    extra_docs_dir = cwd / "docs" / "docsite" / "rst"
    return [
        file
        for file in all_files
        if file.is_relative_to(extra_docs_dir) and file.name.lower().endswith((".rst"))
    ]


def file_of_interest(file: Path) -> bool:
    """
    Determine whether the file is relevant for the docs build.
    """
    if file == Path(CONFIG_FILENAME):
        return True
    if file.is_relative_to("docs/docsite"):
        return True
    if file in (Path("meta/runtime.yml"), Path("galaxy.yml")):
        return True
    if file.is_relative_to("plugins"):
        if file.is_relative_to("plugins/module_utils"):
            return False
        if file.is_relative_to("plugins/plugin_utils"):
            return False
        # Other Python and YAML plugin/ files often contain documentation.
        # Other files do not.
        return file.suffix in (".py", ".yml", ".yaml")
    if file.is_relative_to("roles"):
        # For roles, only files in <role>/meta/ are of interest for documentation.
        return file.match("roles/*/meta/*")
    # So far, no other files are of interest for documentation.
    return False


def add_docs_check(
    *,
    make_docs_check_default: bool = True,
    antsibull_docs_package: PackageTypeOrList = "antsibull-docs",
    ansible_core_package: PackageTypeOrList = "ansible-core",
    validate_collection_refs: t.Literal["self", "dependent", "all"] | None = None,
    extra_collections: list[str] | None = None,
    codeblocks_restrict_types: list[str] | None = None,
    codeblocks_restrict_type_exact_case: bool = True,
    codeblocks_allow_without_type: bool = True,
    codeblocks_allow_literal_blocks: bool = True,
    antsibull_docutils_package: PackageTypeOrList = "antsibull-docutils",
) -> None:
    """
    Add docs-check session for linting.
    """
    run_extra_checks = (
        codeblocks_restrict_types is not None
        or not codeblocks_allow_without_type
        or not codeblocks_allow_literal_blocks
    )

    def compose_dependencies(session: nox.Session) -> list[PackageType]:
        deps = []
        deps.extend(
            check_package_types(
                session,
                "sessions.docs_check.antsibull_docs_package",
                normalize_package_type(antsibull_docs_package),
            )
        )
        deps.extend(
            check_package_types(
                session,
                "sessions.docs_check.ansible_core_package",
                normalize_package_type(ansible_core_package),
            )
        )
        if run_extra_checks:
            deps.extend(
                check_package_types(
                    session,
                    "sessions.docs_check.antsibull_docutils_package",
                    normalize_package_type(antsibull_docutils_package),
                )
            )
        return deps

    def execute_extra_checks(session: nox.Session) -> None:
        all_extra_docs = find_extra_docs_rst_files()
        if not all_extra_docs:
            session.warn(
                "Skipping codeblock checks since no appropriate RST file was found..."
            )
            return
        run_bare_script(
            session,
            "rst-extra",
            use_session_python=True,
            files=all_extra_docs,
            extra_data={
                "codeblocks_restrict_types": codeblocks_restrict_types,
                "codeblocks_restrict_type_exact_case": codeblocks_restrict_type_exact_case,
                "codeblocks_allow_without_type": codeblocks_allow_without_type,
                "codeblocks_allow_literal_blocks": codeblocks_allow_literal_blocks,
            },
            with_cd=True,
            process_messages=True,
        )

    def execute_antsibull_docs(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        changes = get_changes()
        if changes is not None:
            if not any(file_of_interest(file) for file in changes):
                session.warn(
                    "Skip antsibull-docs lint-collection-docs (no files to process)"
                )
                return
        antsibull_docs_version = get_package_version(session, "antsibull-docs")
        if antsibull_docs_version is not None:
            session.log(f"Detected antsibull-docs version {antsibull_docs_version}")
        with session.chdir(prepared_collections.current_path):
            collections_path = f"{prepared_collections.current_place}"
            command = [
                "antsibull-docs",
                "lint-collection-docs",
                "--plugin-docs",
                "--skip-rstcheck",
                ".",
            ]
            if validate_collection_refs:
                command.extend(["--validate-collection-refs", validate_collection_refs])
            if is_new_enough(antsibull_docs_version, min_version="2.18.0"):
                command.append("--check-extra-docs-refs")

            env = {"ANSIBLE_COLLECTIONS_PATH": collections_path}
            if is_new_enough(antsibull_docs_version, min_version="2.24.0"):
                command.extend(["--message-format", "json"])
                with silence_run_verbosity():
                    output = session.run(
                        *command,
                        env=env,
                        silent=True,
                        success_codes=(0, 3),
                    )

                if output:
                    print_messages(
                        session=session,
                        messages=parse_antsibull_docs_errors(
                            output=output,
                        ),
                        fail_msg="antsibull-docs lint-collection-docs failed",
                    )
            else:
                session.run(*command, env=env)

    def docs_check(session: nox.Session) -> None:
        install(session, *compose_dependencies(session))
        prepared_collections = prepare_collections(
            session,
            install_in_site_packages=False,
            extra_collections=extra_collections,
            install_out_of_tree=True,
        )
        if run_extra_checks:
            execute_extra_checks(session)
        if not prepared_collections:
            session.warn("Skipping antsibull-docs...")
        if prepared_collections:
            execute_antsibull_docs(session, prepared_collections)

    docs_check.__doc__ = "Run 'antsibull-docs lint-collection-docs'"
    nox.session(
        name="docs-check",
        default=make_docs_check_default,
    )(docs_check)


__all__ = [
    "add_docs_check",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Change detection specific code.
"""

from __future__ import annotations

import functools
import os
from pathlib import Path

from .config import VCS as VCSConfig
from .config import Config
from .paths.utils import relative_to_walk_up
from .vcs import VCS, VcsProvider
from .vcs.factory import get_vcs_provider

_ENABLE_CD_ENV_VAR = "ANTSIBULL_CHANGE_DETECTION"
_BASE_BRANCH_ENV_VAR = "ANTSIBULL_BASE_BRANCH"


class _CDConfig:
    vcs: VCS
    provider: VcsProvider
    repo: Path
    base_branch: str
    config_dir_is_repo_dir: bool

    def __init__(self, *, config_path: Path, vcs_config: VCSConfig) -> None:
        self.vcs = vcs_config.vcs
        self.provider = get_vcs_provider(self.vcs)
        path = config_path.parent
        repo = self.provider.find_repo_path(path=path)
        if repo is None:
            raise ValueError(f"Cannot find {self.vcs} repository for {path}")
        self.repo = repo
        self.config_dir_is_repo_dir = repo.absolute() == config_path.parent.absolute()
        self.base_branch = vcs_config.development_branch
        env_base_branch = os.environ.get(_BASE_BRANCH_ENV_VAR)
        if env_base_branch:
            self.base_branch = env_base_branch


_cd_initialized = False  # pylint: disable=invalid-name
_cd_config: _CDConfig | None = None  # pylint: disable=invalid-name


def init_cd(
    *,
    config: Config,
    config_path: Path,
    force: bool = False,
    ignore_previous_calls: bool = False,
) -> None:
    """
    Initialize data structures so that the other change detection
    functionality works.
    """
    # We want global context due to the way nox works.
    global _cd_initialized, _cd_config  # pylint: disable=global-statement

    if _cd_initialized and not ignore_previous_calls:
        raise ValueError("init_cd() has already been called!")

    if config.vcs is None:
        _cd_config = None
        _cd_initialized = True
        return

    if not force:
        value = (os.environ.get(_ENABLE_CD_ENV_VAR) or "").lower()
        if value != "true":
            _cd_config = None
            _cd_initialized = True
            return

    _cd_config = _CDConfig(config_path=config_path, vcs_config=config.vcs)
    _cd_initialized = True


def _check_initialized() -> None:
    if not _cd_initialized:
        raise RuntimeError("Internal error: init_cd() has not been called!")


def supports_cd() -> bool:
    """
    Determines whether a antsibull-nox configuration supports CD.
    """
    _check_initialized()
    return _cd_config is not None


def get_vcs_name() -> VCS | None:
    """
    Provide the configured CVS for change detection.

    Returns ``None`` if ``supports_cd() == False``.
    """
    _check_initialized()
    return _cd_config.vcs if _cd_config else None


def get_base_branch() -> str | None:
    """
    Provide the base branch for change detection.

    Returns ``None`` if ``supports_cd() == False``.
    """
    _check_initialized()
    return _cd_config.base_branch if _cd_config else None


def is_config_dir_the_repo_dir() -> bool | None:
    """
    Figure out whether the config dir (the directory containing
    ``antsibull-nox.toml``) is the repository's root directory.

    Returns ``None`` if ``supports_cd() == False``.
    """
    _check_initialized()
    return _cd_config.config_dir_is_repo_dir if _cd_config else None


@functools.cache
def get_changes(*, relative_to: Path | None = None) -> list[Path] | None:
    """
    Acquire a list of changes.

    Returns ``None`` if change detection is not available.

    Returned paths are relative to ``relative_to``, or CWD if ``relative_to is None``.
    """
    _check_initialized()
    cd_config = _cd_config
    if not cd_config:
        return None
    changes = cd_config.provider.get_changes_compared_to(
        repo=cd_config.repo, branch=cd_config.base_branch
    )

    if relative_to is None:
        relative_to = Path.cwd()

    return [relative_to_walk_up(cd_config.repo / path, relative_to) for path in changes]


__all__ = (
    "supports_cd",
    "init_cd",
    "get_base_branch",
    "get_changes",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Config file schema.
"""

from __future__ import annotations

import typing as t

import pydantic as p
from antsibull_fileutils.yaml import load_yaml_file
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version as PackagingVersion

from .ansible import MinPythonVersionConstantsWOATC
from .utils import Version


class _BaseModel(p.BaseModel):
    model_config = p.ConfigDict(frozen=True, extra="allow", validate_default=True)


class ModulesConfig(_BaseModel):
    """
    Configuration for modules/module_utils.
    """

    # Python versions supported by modules/module_utils
    python_requires: str = "default"


class AnsibleTestConfig(_BaseModel):
    """
    Ansible-test configuration file.

    See https://github.com/ansible/ansible/blob/devel/test/lib/ansible_test/config/config.yml
    for details on the format.
    """

    modules: ModulesConfig = ModulesConfig()


def load_ansible_test_config(
    *, accept_missing: bool = False, ignore_errors: bool = False
) -> AnsibleTestConfig:
    """
    Load collection's ansible-test config file.
    """
    try:
        data = load_yaml_file("tests/config.yml")
    except FileNotFoundError:
        if accept_missing or ignore_errors:
            return AnsibleTestConfig()
        raise

    try:
        return AnsibleTestConfig.model_validate(data)
    except ValueError:
        if ignore_errors:
            return AnsibleTestConfig()
        raise


def get_min_python_version(
    config: AnsibleTestConfig, *, ignore_errors: bool = False
) -> MinPythonVersionConstantsWOATC | t.Callable[[Version], bool]:
    """
    Given an Ansible-test config file, figures out the minimum Python version
    for modules and module utils.
    """
    if config.modules.python_requires in ("default", "controller"):
        return t.cast(MinPythonVersionConstantsWOATC, config.modules.python_requires)

    try:
        spec_set = SpecifierSet(config.modules.python_requires)
    except InvalidSpecifier as exc:
        if ignore_errors:
            return "default"
        raise ValueError(
            f"Invalid specifier set in modules.python_requires: {config.modules.python_requires!r}"
        ) from exc

    return lambda version: PackagingVersion(str(version)) in spec_set



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Path utilities.
"""

from __future__ import annotations

import atexit
import functools
import os
import shutil
import sys
import typing as t
from collections.abc import Sequence
from pathlib import Path

from antsibull_fileutils.copier import Copier, GitCopier
from antsibull_fileutils.tempfile import (
    ansible_mkdtemp,
    find_tempdir,
    is_acceptable_tempdir,
)
from antsibull_fileutils.vcs import detect_vcs, list_git_files


def find_data_directory() -> Path:
    """
    Retrieve the directory for antsibull_nox.data on disk.
    """
    return Path(__file__).parent.parent / "data"


def path_walk(
    path: Path, top_down: bool = True, follow_symlinks: bool = False
) -> t.Generator[tuple[Path, list[str], list[str]]]:
    """
    Shim for Path.walk() that falls back to os.walk().
    """
    if hasattr(path, "walk"):
        yield from path.walk(top_down=top_down, follow_symlinks=follow_symlinks)
        return
    for dirpath, dirnames, filenames in os.walk(
        path, topdown=top_down, followlinks=follow_symlinks
    ):
        yield (Path(dirpath), dirnames, filenames)


@functools.cache
def list_all_files() -> list[Path]:
    """
    List all files of interest in the repository.
    """
    directory = Path.cwd()
    vcs = detect_vcs(directory)
    if vcs == "git":
        return [directory / path.decode("utf-8") for path in list_git_files(directory)]
    result = []
    for root, dirs, files in path_walk(directory, top_down=True):
        for file in files:
            result.append(root / file)
        if root == directory and ".nox" in dirs:
            dirs.remove(".nox")
    return result


def remove_path(path: Path) -> None:
    """
    Delete a path.
    """
    if not path.is_symlink() and path.is_dir():
        shutil.rmtree(path)
    elif path.exists():
        path.unlink()


def copy_collection(
    source: Path, destination: Path, *, copy_repo_structure: bool = False
) -> None:
    """
    Copy a collection from source to destination.

    Automatically detect supported VCSs and use their information to avoid
    copying ignored files.
    """
    if destination.exists():
        remove_path(destination)
    vcs = detect_vcs(source)
    copier: Copier
    if vcs == "git":
        copier = GitCopier(copy_repo_structure=copy_repo_structure)
    else:
        copier = Copier()
    copier.copy(source, destination, exclude_root=[".nox", ".tox"])


def _is_acceptable_tempdir(directory: Path, excludes: Sequence[Path]) -> bool:
    # Avoid being inside an ansible_collections tree
    if not is_acceptable_tempdir(directory):
        return False
    # Avoid something that's inside one of the excludes
    for exclude in excludes:
        if directory.is_relative_to(exclude):
            return False
    return True


def get_outside_temp_directory(exclude: Sequence[Path] | Path | None = None) -> Path:
    """
    Find a temporary directory root that's outside the provided path.

    Raises ``ValueError`` in case no candidate was found.
    """
    if exclude is None:
        exclude = [Path.cwd()]
    elif isinstance(exclude, Path):
        exclude = [exclude]
    return find_tempdir(lambda path: _is_acceptable_tempdir(path, exclude))


def create_temp_directory(basename: str) -> Path:
    """
    Create a temporary directory outside an existing ansible_collections tree.
    """
    path = ansible_mkdtemp(prefix=basename)

    def cleanup() -> None:
        remove_path(path)

    atexit.register(cleanup)
    return path


def copy_directory_tree_into(source: Path, destination: Path) -> None:
    """
    Copy the directory tree from ``source`` into the tree at ``destination``.

    If ``destination`` does not yet exist, it will be created first.
    """
    if not source.is_dir():
        return
    destination.mkdir(parents=True, exist_ok=True)
    for root, _, files in path_walk(source):
        path = destination / root.relative_to(source)
        path.mkdir(exist_ok=True)
        for file in files:
            dest = path / file
            remove_path(dest)
            shutil.copy2(root / file, dest, follow_symlinks=False)


def relative_to_walk_up(path: Path, relative_to: Path) -> Path:
    """
    Path.relative_to()'s walk_up kwarg was only added in Python 3.12.
    This function provides a compatibility shim for older Python versions.
    """
    if sys.version_info < (3, 12):
        return Path(os.path.relpath(path, relative_to))
    return path.relative_to(relative_to, walk_up=True)


__all__ = (
    "copy_collection",
    "copy_directory_tree_into",
    "create_temp_directory",
    "find_data_directory",
    "get_outside_temp_directory",
    "list_all_files",
    "path_walk",
    "relative_to_walk_up",
    "remove_path",
)
